# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

__author__ = "Sqreen"
__copyright__ = "Copyright 2016 - 2020, Sqreen"
__email__ = "contact@sqreen.io"
__license__ = "proprietary"
__version__ = "1.21.0b4"
